# pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring, no-self-use
import unittest
from unittest.mock import patch
import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestClassifier
from iris_dataset.modules import retrain_predictor
from .patch_decorators import patch_modulify


metadata = {
    "columns": {
        "X_columns": {
            "sepal_length": {"name": "sepal_length", "type": "float"},
            "sepal_width": {"name": "sepal_width", "type": "float"},
            "petal_length": {"name": "petal_length", "type": "float"},
            "petal_width": {"name": "petal_width", "type": "float"},
        },
        "y_columns": {"species": {"name": "species", "type": "str"}},
    }
}


class InputDS:
    def __init__(self, data, metadata_ref=None):
        self.data = data
        self.metadata = metadata_ref if metadata_ref is not None else {}
        self.iter = iter([d[1] for d in self.data.iterrows()])

    def __iter__(self):
        return self.iter
        # self.iter = iter([d[1] for d in self.data.iterrows()])
        # return self

    def __next__(self):
        return next(self.iter)

    def get_data(self):
        return self.data


# This test is buggy
# class TestModulifiedRetrainPredictor(unittest.TestCase):
#     def test_output_types(self):
#         # This is a bit hacky but in tests it's fine.
#         output_types_fn = retrain_predictor().retrain_predictor.__closure__[-1].cell_contents
#         assert output_types_fn() == [RandomForestModel]


@patch_modulify(retrain_predictor)
class TestRetrainPredictor(unittest.TestCase):
    @patch("iris_dataset.modules.retrain_predictor.RandomForestModel")
    def test_retrain_predictor(self, mocked_rfmodel):
        np.random.seed(1111)
        mocked_rfmodel.return_value = "ok"
        fixture = pd.read_csv("./data/iris.csv")
        dataset = InputDS(fixture, metadata)  # Mock DS??
        [result] = retrain_predictor.retrain_predictor(dataset)
        assert result == "ok"
        assert len(mocked_rfmodel.call_args_list) == 1
        args, _kwargs = mocked_rfmodel.call_args_list[0]
        assert len(args) == 2
        assert isinstance(args[0], RandomForestClassifier)
        assert args[0].oob_score
        assert args[1] == {
            "oob_score": 0.9466666666666667,
            "error": 0.053333333333333344,
            "x_labels": metadata["columns"]["X_columns"],
            "y_labels": {"species": {"name": "species", "type": "str"}},
        }



"""
    DataStructure to serialize PredictionsCombinations
    It stores the data to ElasticSearch.
"""

from datetime import datetime
from soil.data_structures.es_data_structure import ESDataStructure


def _row_2_action(index, _metadata):
    def action_fn(index_row):
        _i, row = index_row
        return {"_index": index, "generated_at": datetime.now(), **row.to_dict()}

    return action_fn


class Predictions(ESDataStructure):
    """
    DataStructure to serialize PredictionsCombinations
    It stores the data to ElasticSearch.
    """

    def _row_to_action(self, *args):
        # pylint: disable=no-self-use
        return _row_2_action(*args)



"""
This module overrides Python's import system
https://docs.python.org/3/library/importlib.html
https://stackoverflow.com/a/43573798/3481480
"""


import sys
import os
from typing import Optional, Sequence, Union
from importlib.machinery import ModuleSpec
from importlib.abc import MetaPathFinder
from importlib.util import spec_from_file_location
import types
import soil as _soil  # NOQA pylint:disable=unused-import

UNITTEST_MOCKS_FOLDER = "test/unit/unittest_mocks"


class Finder(MetaPathFinder):
    # pylint:disable=no-self-use
    """
    Load files that should come from the DB from unittest_mocks
    """

    def find_spec(
        self,
        fullname: str,
        path: Optional[Sequence[Union[bytes, str]]],
        _target: Optional[types.ModuleType] = None,
    ) -> Optional[ModuleSpec]:
        """Find the module in UNITTEST_MOCKS_FOLDER"""
        if "." in fullname:
            *parents, name = fullname.split(".")
        else:
            name = fullname
            parents = []

        if not (
            len(parents) > 1
            and parents[0] == "soil"
            and (parents[1] == "modules" or parents[1] == "data_structures")
        ):
            # If the module is not from soil.modules or soil.data_structures use the default loader
            return None

        path = UNITTEST_MOCKS_FOLDER
        folder = parents[2:]
        filename = os.path.abspath(os.path.join(path, *folder, f"{name}.py"))

        if os.path.exists(filename):
            return spec_from_file_location(fullname, filename)
        return None


def install() -> None:
    """Inserts the finder into the import machinery"""
    sys.meta_path.insert(0, Finder())



""" Module to train a predictor. """
import argparse
import logging
import soil
from soil import logger
from soil.modules.retrain_predictor import retrain_predictor

# These lines allow the use the use of data structures in its current state
# Use these during code development if any change has been done since data was uploaded
# from soil.data_structures import es_data_structure
# from soil.modules.to_es_data_structure import to_es_data_structure


logger.setLevel(logging.DEBUG)


def train_predictor():
    """Main function to train a predictor."""

    data = soil.data("iris-data")
    predictor_ref, = retrain_predictor(data)
    soil.alias("trained-model-simple", predictor_ref)


def main():
    """Argument parsing."""
    parser = argparse.ArgumentParser()
    args = parser.parse_args()
    train_predictor(**vars(args))


if __name__ == "__main__":
    main()



""" This program runs when new data arrives. """
import logging
import argparse
import json
import soil
from soil import logger
from soil.modules.to_es_data_structure import to_es_data_structure
from ..lib.utils import get_columns_dictnames, read_data

logger.setLevel(logging.DEBUG)
# logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')


def _new_data(input_file=None, schema_file=None):

    # Read schema
    schema = json.load(schema_file)
    columns = get_columns_dictnames(schema["columns"])

    # Read data
    # Adding new data does not delete previous data unless metadata["rewrite"]= True
    # (those entries whose id appear in both datasets are rewritten in any case)

    # Read data as a list of dictionaries
    data = read_data(input_file)

    soil_data_structure = soil.data(data, metadata={"columns": columns, "index": "iris-data"})
    # Add "rewrite":True to metadata to replace old data

    # Upload a list of dictionaries (without NaN values) to elasticsearch
    elasticsearch_data_structure, = to_es_data_structure(soil_data_structure)
    soil.alias("iris-data", elasticsearch_data_structure, roles=["user"])


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # inputs given by soil.yml must start by --
    # inputs given during script call (soil run ... <input-file>) do not
    parser.add_argument(
        "input_file", type=argparse.FileType("r"), help="The input file"
    )
    parser.add_argument(
        "--schema-file",
        type=argparse.FileType("r"),
        help="The file with the schema of the data",
        required=True,
    )
    args = parser.parse_args()
    _new_data(**vars(args))



""" Module to train a predictor. """
import argparse
import logging
import soil
import pandas as pd
from soil.modules.make_predictions import make_predictions

soil.logger.setLevel(logging.INFO)


def get_preditictions(unlabeled_data_file=None):
    """Main function to train a predictor."""

    model = soil.data("trained-model-simple")

    data = pd.read_csv(unlabeled_data_file)
    data_ref = soil.data(data)

    preds, = make_predictions(model, data_ref)

    print("The predicted species for the input data are the following:")
    for i in preds.data:
        print(i)

    soil.alias("preds-simple", preds)


def main():
    """Argument parsing."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--unlabeled-data-file",
        type=argparse.FileType("r"),
        help="Pass unlabeled data here",
        required=True,
    )
    args = parser.parse_args()
    get_preditictions(**vars(args))


if __name__ == "__main__":
    main()
